local L = AceLibrary("AceLocale-2.2"):new("BananaDKP-Symphony")
local roster = AceLibrary("Roster-2.1")

local counterRunning = nil

L:RegisterTranslations("enUS", function() return {
	["Start"] = true,
	["Pause"] = true,
	["Stop"] = true,
	["Start the automatic hour counter."] = true,
	["Pause the automatic hour counter."] = true,
	["Stop the automatic hour counter."] = true,

	["Automatic hour counting has started, the first increment will be in ~%d minutes."] = true,
	["Automatic hour counting paused, ~%d minutes to next update."] = true,
	["Automatic hour counting stopped, ~%d minutes to next update."] = true,
	["Automatic hour counting resumed."] = true,

	["Automatic"] = true,
	["Manual"] = true,

	["Increment"] = true,
	["Increment the hour counter for a player or the entire raid+sitout list (*) by one."] = true,
	["<player or *>"] = true,

	["Decrement"] = true,
	["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."] = true,

	["The hour counter has been incremented by 1 for %s."] = true,
	["The hour counter has been decremented by 1 for %s."] = true,
	["The hour counter has been incremented by 1 for all players in the raid and on the sitout list."] = true,

	["No items looted by %s"] = true,
	["Items looted by %s: %s"] = true,
	["%s has raided %d hours"] = true,

	["%s looted %s, do you want to add it to his loot table?"] = true,
	["%s looted %s, it has been added to his item table."] = true,

	["bdkp items <player>; list of the items the player has looted"] = true,
	["bdkp hours <player>; number of hours the player has raided"] = true,

	["Add"] = true,
	["Ignore"] = true,
} end)

local lootDialogOpen = nil
local lootQueue = {}

local halfHour = 60 * 30
local oneHour = halfHour * 2

local paused = nil
local counterRunning = nil

local db = nil

local _G = getfenv(0)

Symphony = BananaDKP:NewModule("Symphony")

Symphony.defaultDatabase = {
	databaseVersion = 2, -- Database version.
	hours = {}, -- Key: Player, Value: Hours of raiding, autoincremented every 30min.
	items = {}, -- Key: Player, Value: Table of item ids looted.
	lastUpdate = {}, -- Key: Player, Value: Time of last update for this player.
	hourTickerSavePoint = 0,
}

Symphony.options = {
	autoHeader = {
		type = "header",
		name = L["Automatic"],
		order = 100,
	},
	start = {
		type = "execute",
		name = L["Start"],
		desc = L["Start the automatic hour counter."],
		func = function() Symphony:StartCounter() end,
		order = 101,
		disabled = function() return not UnitInRaid("player") or counterRunning end,
	},
	pause = {
		type = "toggle",
		name = L["Pause"],
		desc = L["Pause the automatic hour counter."],
		get = function() return paused and true or false end,
		set = function(v) Symphony:PauseCounter(v) end,
		order = 102,
		disabled = function() return not counterRunning end,
	},
	stop = {
		type = "execute",
		name = L["Stop"],
		desc = L["Stop the automatic hour counter."],
		func = function() Symphony:StopCounter() end,
		disabled = function() return not counterRunning end,
		order = 103,
	},
	spacer = {
		type = "header",
		name = " ",
		order = 150,
	},
	manualHeader = {
		type = "header",
		name = L["Manual"],
		order = 200,
	},
	plusone = {
		type = "text",
		name = L["Increment"],
		desc = L["Increment the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = function(player) Symphony:IncrementHourCounter(player) end,
		order = 201,
	},
	minusone = {
		type = "text",
		name = L["Decrement"],
		desc = L["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = function(player) Symphony:DecrementHourCounter(player) end,
		order = 202,
	},
}

function Symphony:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end

	if type(popup["SymphonyItemQuery"]) ~= "table" then
		popup["SymphonyItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnAccept = function(data)
				Symphony:AddItemToItemTable(unpack(data))
			end,
			OnCancel = function()
				Symphony:IgnoreItem()
			end,
		}
	end

	db = self.db.profile

	self.core:RegisterWhisperHandler("items", function(event, player)
		player = player:lower()
		if db.items[player] then
			return string.format(L["Items looted by %s: %s"], player, table.concat(db.items[player], ", "))
		else
			return string.format(L["No items looted by %s"], player)
		end
	end, L["bdkp items <player>; list of the items the player has looted"])

	self.core:RegisterWhisperHandler("hours", function(event, player)
		player = player:lower()
		return string.format(L["%s has raided %d hours"], player, db.hours[player] or 0)
	end, L["bdkp hours <player>; number of hours the player has raided"])

	self:RegisterEvent("BananaDKP_LeftRaid", "StopCounter")
end

-- Every minute we save our current timer, so that if we get disconnected or
-- something, we know where to start from next time.
function Symphony:RepeatedTimeDBStore()
	if not counterRunning then return end
	local timeLeft
	if paused then
		timeLeft = paused
	else
		timeLeft = select(2, self:IsEventScheduled("Ticker"))
	end
	db.hourTickerSavePoint = timeLeft
	self:ScheduleEvent("TickerStore", self.RepeatedTimeDBStore, 60, self)
end

function Symphony:StartCounter()
	assert(not counterRunning, "Counter is already running.")
	local time = 0
	if db.hourTickerSavePoint > 0 then
		time = db.hourTickerSavePoint
	else
		time = halfHour
	end
	self:ScheduleEvent("Ticker", self.IncrementHourCounter, time, self)
	counterRunning = true
	self:RepeatedTimeDBStore()
	local minutes = math.floor(time / 60)
	self.core:Print(L["Automatic hour counting has started, the first increment will be in ~%d minutes."]:format(minutes))
end

function Symphony:PauseCounter(input)
	assert(counterRunning, "Can't pause something that is not running.")
	if input then
		paused = select(2, self:IsEventScheduled("Ticker"))
		self:CancelScheduledEvent("Ticker")
		local minutes = math.floor(paused / 60)
		self.core:Print(string.format(L["Automatic hour counting paused, ~%d minutes to next update."], minutes))
	else
		self:ScheduleEvent("Ticker", self.IncrementHourCounter, paused, self)
		paused = nil
		self.core:Print(L["Automatic hour counting resumed."])
	end
end

function Symphony:StopCounter()
	if not counterRunning then return end
	local timeLeft = select(2, self:IsEventScheduled("Ticker"))
	if not timeLeft then timeLeft = paused end
	self:CancelScheduledEvent("Ticker")
	self:CancelScheduledEvent("TickerStore")
	counterRunning = nil
	timeLeft = math.floor(timeLeft / 60)
	self.core:Print(string.format(L["Automatic hour counting stopped, ~%d minutes to next update."], timeLeft or -1))
	paused = nil
	db.hourTickerSavePoint = 0
end

function Symphony:ItemLooted(player, itemLink, itemRarity)
	if itemRarity < 4 then return end
	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink, itemRarity})
	else
		lootDialogOpen = true
		StaticPopupDialogs["SymphonyItemQuery"].text = string.format(L["%s looted %s, do you want to add it to his loot table?"], player, itemLink)
		local popup = StaticPopup_Show("SymphonyItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end

function Symphony:AddItemToItemTable(player, itemLink)
	lootDialogOpen = nil

	player = player:lower()
	if not db.items[player] then db.items[player] = {} end
	table.insert(db.items[player], itemLink)
	db.lastUpdate[player] = time()

	self.core:Print(string.format(L["%s looted %s, it has been added to his item table."], player, itemLink))

	if #lootQueue > 0 then
		self:ItemLooted(unpack(table.remove(lootQueue, 1)))
	else
		self:TriggerEvent("BananaDKP_Synchronize")
	end
end

function Symphony:IgnoreItem()
	lootDialogOpen = nil

	if #lootQueue > 0 then
		self:ItemLooted(unpack(table.remove(lootQueue, 1)))
	end
end

function Symphony:IncrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:lower()
		if db.hours[input] then
			db.hours[input] = db.hours[input] + 1
		else
			db.hours[input] = 1
		end
		self.core:Print(string.format(L["The hour counter has been incremented by 1 for %s."], input))
	else
		if type(input) == "string" and input ~= "*" then
			error("Can't increment the hour counter for "..input..".")
			return
		end
		for n, u in pairs(roster.roster) do
			if u.class ~= "PET" then
				local k = n:lower()
				if not db.hours[k] then
					db.hours[k] = 1
				else
					db.hours[k] = db.hours[k] + 1
				end
				db.lastUpdate[k] = time()
			end
		end
		for k, v in pairs(BananaDKP.db.profile.sitouts) do
			local key = k:lower()
			if not db.hours[key] then
				db.hours[key] = 1
			else
				db.hours[key] = db.hours[key] + 1
			end
			db.lastUpdate[key] = time()
		end
		self.core:Print(L["The hour counter has been incremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			self:ScheduleEvent("Ticker", self.IncrementHourCounter, oneHour, self)
		end
	end

	self:TriggerEvent("BananaDKP_Synchronize")
end

function Symphony:DecrementHourCounter(player)
	assert(type(player) == "string", "Invalid input.")
	player = player:lower()
	if not db.hours[player] then
		self.core:Print(string.format("No such player: %s.", player))
	else
		db.hours[player] = db.hours[player] - 1
		self.core:Print(string.format(L["The hour counter has been decremented by 1 for %s."], player))
	end
end

-- DKP Formula:
-- #hoursRaided / (1 + #itemsLooted)
--
-- Pretty simple, I know :P That's what my guild wanted, so that's what I did.

-- Thanks, Wobin.
local function round(n,shift)
	shift = 10^shift
	return math.floor((n*shift)+0.5)/shift
end

function Symphony:GetDKP(name)
	if type(name) ~= "string" then return nil end
	name = name:lower()
	if not db.hours[name] then return nil end
	local numItems = (db.items[name] and #db.items[name]) or 0
	return round(db.hours[name] / (1 + numItems), 2)
end

function Symphony:GetLootedItems(name)
	return db.items[name:lower()]
end

local function print(t)
	BananaDKP:Print(t)
end

function Symphony:SynchronizeDKP(sender, lastUpdate, hours, items)
	local toUpdate = {}

	-- Loop the lastUpdate table and see if anyone needs to be updated.
	for k, v in pairs(lastUpdate) do
		if not db.lastUpdate[k] or lastUpdate[k] > db.lastUpdate[k] then
			table.insert(toUpdate, k)
		end
	end

	-- Loop the toUpdate table and sync our aliases, raids and items with the
	-- members listed.
	for i, v in ipairs(toUpdate) do
		db.lastUpdate[v] = lastUpdate[v]
		db.hours[v] = hours[v]
		db.items[v] = {}
		for _, item in ipairs(items[v]) do
			table.insert(db.items[v], item)
		end
	end
end

function Symphony:GetDKPSynchronizationData(sender)
	return db.lastUpdate, db.hours, db.items
end

function Symphony:GetDBVersion()
	return db.databaseVersion
end

